/*M*
//                    INTEL CORPORATION PROPRIETARY INFORMATION
//      This software is supplied under the terms of a license agreement or
//      nondisclosure agreement with Intel Corporation and may not be copied
//      or disclosed except in accordance with the terms of that agreement.
//            Copyright (c) 1999 Intel Corporation.  All Rights Reserved.
//
//  PVCS:
//      $Workfile:   imb_lin.c  $
//      $Revision:   1.3  $
//      $Modtime:   14 Oct 1999 13:03:28  $
//      $Author:   mramacha  $
//		$Log:   //cospublic004/shareddata/ISC/ISC2x/LNWORK/SRC/imb/imb_lin.c.v  $
 * 
 *    Rev 1.3   14 Oct 1999 12:58:28   mramacha
 * Changes to mmap.
 * 
 *    Rev 1.2   09 Sep 1999 13:25:48   mramacha
 * Changes to ioctl and mmap
 * 
 *    Rev 1.1   Aug 29 1999 20:38:18   mramachx
 * Code cleanup and buf fixes.
 * 
 *    Rev 1.0   Aug 27 1999 16:55:50   mramachx
 * Initial Linux Checkin
//
//  Purpose:    This file contains the Linux driver entry points as well as 
//				Linux specific code for the IMB driver.
//
//
//  Contents:
// 			init_module
// 			cleanup_module
// 			IMBopen
// 			IMBclose
// 			IMBioctl
// 			IMBmmap 
//			DriverEntry
//			DriverExit
//			ImbShutdown
//
//
//  Notes:
//			The driver in Linux (as in Unixware and Solaris) can be opened
//			by only one user-level app since there is only one device entry.
//			Hence we do not worry about maintaining per-device state. But this
//			can be modified fairly easily. MR
//
*M*/

/* Note that linux/module.h must be included only here and not the other
 * files. Hence it is not in imb_os.h. Mahendra
 */
#include <linux/module.h>
#include "imb_os.h"
#include "imb_drv.h"



/* Note that a major of 0 means the OS can dynamically asign a major number 
 * for our device. Our startup script in /etc/rc.d/init.d/isc must ensure
 * that we clean up stale /dev entries and create new ones based on the
 * Major number assigned with our module was loaded.
**/
#define	DEFAULT_IMB_MAJOR 0 
#define IMB_DRV_NAME "imb"

/////////////////////////////////////////////////////////////////
//  prototypes for driver entry points
/////////////////////////////////////////////////////////////////
static int 	 IMB_open(struct inode *, struct file *);
static int  IMB_release(struct inode * , struct file * );
static int 	 IMB_ioctl(struct inode *, struct file *, unsigned int , unsigned long );

static int  IMBmmap(struct file *, struct vm_area_struct *);

////////////////////////////////////////////////////////////////
// other function prototypes
////////////////////////////////////////////////////////////////
static int DriverEntry( PDEVICE_EXTENSION deviceExtension);
static void DriverExit(PDEVICE_EXTENSION deviceExtension);
static NTSTATUS ImbShutdown ( PDEVICE_EXTENSION deviceExtension);
extern void TimerDpcFunc(unsigned long arg);

/////////////////////////////////////////////////////////////////////////////
// The file_operations defines the driver entry points we support in imb.
// It is used by init_module() to register the driver. Note that I have
// defined IMB_fops in a way supported by gcc compiler. This way I don't
// have to worry if the struture is changed in future releases.
//
// If you need to compile using a different compiler, you may need to ifdef
// the declaration for gcc and in the else put the complete structure with
// NULLs for the other fields.
/////////////////////////////////////////////////////////////////////////////
static struct file_operations    IMB_fops = {
    open:		IMB_open,
    release:	IMB_release,
	ioctl:		IMB_ioctl,
    mmap:	    IMBmmap
};

MODULE_AUTHOR("Mahendra Ramachandran, Intel Corp.");
MODULE_DESCRIPTION("driver for communicating with f/w for instrumentation");

//////////////////////////////////////////////////////////
// A couple of strictly local types
//////////////////////////////////////////////////////////
// structure used when managing ISA property "3-tuples"
typedef struct isa_reg {
    uint32_t    type;
    uint32_t    addr;
    uint32_t    size;
} isa_reg_t;

// Size field used to determine size of table mapping is defined in DMI BIOS
// spec to be a word type, or 16 bits wide.  It's implemented in dmibio.h
// as an unsigned short (DMI_BIOS_ENTRY_POINT_STRUCT.tableLength).  Hence,
// the max is 65535.  Need to pin down a max because you can't get the size
// requested from the args to _mmap().
#define MAX_REQ_MAP_SIZE    65535


//
// These two defines are used to signal success or failure between functions
// They are similar to DDI_SUCCESS and DDI_FAILURE in DDI/DKI compliant
// unixes. The driver entry points will use errno values to signal failure 
// to the OS and user-level code.

#define SUCCESS 0
#define FAILURE -1

///////////////////////////////////////////////////////////
// Global state variable
//////////////////////////////////////////////////////////
PDEVICE_EXTENSION   deviceExtension;
int imb_major;

///////////////////////////////////////////////////////////
//   Function Definitions begin Here
///////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
//  init_module
///////////////////////////////////////////////////////////
/*F*
//  Name:       init_module
//  Purpose:    supports loading of the driver
//  Context:    called by system when loading the driver as a 
//				result of the insmod command
//  Returns:    0 on success, negative number on failure
//  Parameters: none
//  Notes:      none
*F*/
int
init_module(void)
{

	//
	//Register with the kernel our loadable module
	//
    imb_major = register_chrdev(DEFAULT_IMB_MAJOR, IMB_DRV_NAME, &IMB_fops);
	DEBUG("imb: register_chrdev returned imb_major %d\n", imb_major);

	//
    //Since DEFAULT_IMB_MAJOR = 0, ie let the kernel assign a major number,
	// the return value must be > 0.
	//If we had given a specific major number, then the return 
	//should be 0 to indicate success.
	//
    if (imb_major <= 0) { 
		printk(KERN_ERR "imb: Unable to register_chrdev\n");
		return (-EBUSY);
	}

	//
	// Now allocate the space for the global data structure.
	//
	
    deviceExtension = (PDEVICE_EXTENSION)kmalloc(
											sizeof(DEVICE_EXTENSION), 
											GFP_KERNEL);
	memset(deviceExtension, sizeof(DEVICE_EXTENSION), 0);

	DEBUG("imb allocated deviceExtension addr 0x%x\n", deviceExtension);

    if (DriverEntry(deviceExtension) == FAILURE) {
		printk(KERN_ERR "IMB: DriverEntry failure. unregister device\n");
		kfree(deviceExtension);
		unregister_chrdev(imb_major, IMB_DRV_NAME);
		return(-EIO);
	}

	DEBUG("imb: init_module successful\n");

    return(SUCCESS);
} //init_module

//////////////////////////////////////////////////////////
//  cleanup_module
//////////////////////////////////////////////////////////
/*F*
//  Name:       cleanup_module
//  Purpose:    supports unloading of the driver
//  Context:    called to unload the driver - remmod command
//  Returns:    zero on success, negative number on failure
//  Parameters: none
//  Notes:      none
*F*/
void
cleanup_module(void)
{
	DEBUG("imb cleanup_module deviceExtension addr 0x%x\n", deviceExtension);
    DriverExit(deviceExtension);
	kfree(deviceExtension);

	//Tell the kernel to dis-engage us. We are going bye-bye.
	//
    if (unregister_chrdev(imb_major, IMB_DRV_NAME) < 0)
		printk(KERN_ERR "IMB: unable to unregister_chrdev\n");


	DEBUG("imb: cleanup_module says bye-bye\n");
} //cleanup_module

/////////////////////////////////////////////////////////////////
// DriverEntry
/////////////////////////////////////////////////////////////////
/*F*
//  Name:       DriverEntry
//  Purpose:    Initialize the deviceExtension state structure
//  Context:    Called by init_module routine
//  Returns:    0 on success and a negative error number on failure.
//  Parameters:
//              deviceExtension -- state structure for driver
//  Notes:      none
*F*/


static int
DriverEntry( PDEVICE_EXTENSION deviceExtension)
{
    ImbInterface 			**imbIf;
    extern ImbInterface 	*ImbInterfaces[];

	DEBUG("imb DriverEntry. deviceExtension = 0x%x \n", deviceExtension);

    // Check to see if we have an interface that we recognize
    //
    for( imbIf = ImbInterfaces ; *imbIf != NULL ; imbIf++ ) {
        if( (*imbIf)->ifProbe() == TRUE ) {
			DEBUG("IMB DriverEntry: Probe success\n");
            break;
		}
    }
    if( *imbIf == NULL ) {
        printk(KERN_ERR "DriverEntry: ifProbe failed\n");
		return(FAILURE);
    }

    //
    // Set up the device extension.
    //
    deviceExtension->unloadingDriver    = 0;

	//Set up the mesgTimer and the TimerDpcFunc (similar to KeInitializeTimer
	//and KeInitializeDpc for NT and unlike timeout in Unix). 
	// The expires field will get set when starting timer

	DEBUG("imb DriverEntry. init_timer for mesgTimer\n");

	init_timer( &deviceExtension->mesgTimer );
	deviceExtension->mesgTimer.data = (unsigned long)deviceExtension;
	deviceExtension->mesgTimer.function = TimerDpcFunc;

	deviceExtension->timeOutFlag		= 0;
    deviceExtension->shutdownCode       = SD_NO_ACTION;
    deviceExtension->imbSeqNo           = 0;
    deviceExtension->imbIf              = *imbIf;
	deviceExtension->asyncQSema         = NULL;


    //
    // initialize the request semaphore
    //
	IMB_LOCK_ALLOC(deviceExtension->requestSemaphor);
	if (deviceExtension->requestSemaphor == NULL) {
		printk(KERN_ERR "imb: DriverEntry - IMB_LOCK_ALLOC failed\n");
		return(FAILURE);
	}

    //
    // Allocate space and initialize the async messaging subsystem
	// Note: kmalloc is gauranteed to succeed as we are using GFP_KERNEL
	// which will let it sleep if necessary to get us the memory.
    //
    deviceExtension->asyncQ.queue = (ImbAsyncEntry *) kmalloc(
									IMB_ASYNC_QUEUE_SIZE * 
									sizeof(ImbAsyncEntry),
									 GFP_KERNEL);

    deviceExtension->asyncQ.queueSize = IMB_ASYNC_QUEUE_SIZE;


    ImbAsyncInit( &deviceExtension->asyncQ );
    
    // assume by default old IPMI version and later detect the actual one.
    deviceExtension->bmcVer =   IPMI_OLD_VER;
    getBmcIpmiVersion(deviceExtension);

	DEBUG("IMB DriverEntry: Allocations etc went well\n");
    return(SUCCESS);
} //DriverEntry

//////////////////////////////////////////////////////////////////
// DriverExit
//////////////////////////////////////////////////////////////////
/*F*
//  Name:       DriverExit
//  Purpose:    OS-independent driver detach/unload routine
//  Context:    called by IMBdetach
//  Returns:    void
//  Parameters: deviceExtension -- driver state structure
//  Notes:      none
*F*/
static void
DriverExit(PDEVICE_EXTENSION deviceExtension)
{
	DEBUG("imb: DriverExit called to free deviceExtension fields\n");

	//Cancel our mesgTimer
    if (timer_pending( &deviceExtension->mesgTimer)) {
		DEBUG("imb DriverExit. mesgTimer timer_pending true so delete\n");
		del_timer( &deviceExtension->mesgTimer);
	}

    //
    //De-allocate all the space we allocated in the DriverEntry
	IMB_LOCK_DEALLOC(deviceExtension->requestSemaphor);

	//
	//First clean up the asyncQ's fields and then the queue itself
	//
	IMB_LOCK_DEALLOC(deviceExtension->asyncQ.syncSem);
	IMB_LOCK_DEALLOC(deviceExtension->asyncQ.waitEvent.plock);

	//Incase we get here and a thread from user-space is still blocked
	if (deviceExtension->asyncQ.waitEvent.waitOn) {
		wake_up_interruptible(&deviceExtension->asyncQ.waitEvent.waitOn);
		deviceExtension->asyncQ.waitEvent.waitOn = NULL;
	}

	if (timer_pending( &deviceExtension->asyncQ.waitEvent.asyncMesgTimer)) {
		DEBUG("imb DriverExit. asyncMesgTimer timer_pending true so delete\n");
		del_timer(&deviceExtension->asyncQ.waitEvent.asyncMesgTimer);
	} 
	kfree(deviceExtension->asyncQ.queue);
	
    return;
} //DriverExit


//////////////////////////////////////////////////////////////////
//  IMB_open
//////////////////////////////////////////////////////////////////
/*F*
//  Name:       IMB_open
//  Purpose:    open the device for use by user
//  Context:    called when user calls open() on device node
//  Returns:    0 on success, an appropriate errno on failure
//  Parameters: device number, flags, open type, credential pointer
//  Notes:      none
*F*/
/*ARGSUSED*/
static int
IMB_open(struct inode *inodep, struct file *filep)
{
	int imb_minor;

	imb_minor = MINOR(inodep->i_rdev);

	// For now we assume that only one minor exists for our driver
	// and we want to validate that assumption. In the future if
	// we support multiple devices, we would make use of this number
	// to associate each open with the proper device extension struct.

	if (imb_minor != 0) {
		printk(KERN_ERR "imb: open failed due to invalid minor device ID\n");
		return(-ENODEV);	//ENODEV: No such device
	}

    //
    // Associate the private_data pointer for this open to the DeviceExtension
	// Note this is not essential since we have only one DeviceExtension
	// that we deal with right now and it is a global. But we can extend
	// this in the future if we need to support multiple devices by using
	// this method. All future access to the driver will use the private_data
	// pointer and not the global structure. mahendra 7/99.
    //
    filep->private_data = deviceExtension ;

    if (filep->private_data == NULL) {
		printk(KERN_ERR "imb: open failed due to invalid device struture\n");
        return(-ENXIO);		//ENXIO: No such device or address
	}

	DEBUG("IMB_open: Minor %d, Device 0x%x\n", imb_minor, filep->private_data);

	MOD_INC_USE_COUNT;
    return(SUCCESS);
} //IMBopen


///////////////////////////////////////////////////////////////////////
// IMB_release
///////////////////////////////////////////////////////////////////////
/*F*
//  Name:       IMB_release
//  Purpose:    closes opened device node
//  Context:    called when user calls close() on open device node
//  Returns:    0 on success, appropriate errno on failure
//  Parameters: device number, flags, type, credential pointer
//  Notes:      none
*F*/
/*ARGSUSED*/
static int
IMB_release(struct inode *inodep, struct file *filep)
{

    //
    // Check the pointer to the device extension.
    //

    if (filep->private_data == NULL)
        return(-ENXIO);		//ENXIO: No such device or address

	//Note we do not free the DeviceExtension since it is our global one
	//and we'll do that when we unload the driver. If we supported multiple
	//devices, we'd have more work to do here for the device that is being
	//closed.
	DEBUG("IMB_release: DeviceExtension 0x%x\n", filep->private_data);

    // De-allocate the DPC syncronization variable so that it can
    // be allocated the next time DPC thread starts up. Otherwise
    // stoppping and starting basebrd will not work. MR
    ImbUnixAsyncDeallocate(filep->private_data);


	MOD_DEC_USE_COUNT; //If the module inuse count went to 0, driver will
					   //be unloaded.
    return(SUCCESS);
} //IMB_release


///////////////////////////////////////////////////////////////////////
// IMB_ioctl
///////////////////////////////////////////////////////////////////////
/*F*
//  Name:       IMB_ioctl
//  Purpose:    user interface to driver
//  Context:    called when user calls ioctl() on open device node
//  Returns:    0 on success, errno on failure, error number on
//              recoverable error
//  Parameters: Compiles with the kernel driver entry prototype.
//  Notes:		No locks held since only one ioctl can be outstanding at a time
//				Multi-threaded apps like cif are sync'ed in libimbapi.
//			
//             
*F*/

/*ARGSUSED*/
static int
IMB_ioctl(struct inode *inodep, struct file *filep, unsigned int cmd, 
			unsigned long arg)
{
    PDEVICE_EXTENSION deviceExtensionP = filep->private_data;
    long status = STATUS_SUCCESS;		
	DWORD inputBufferLength, outputBufferLength, length;	
    struct smi  smi,*smi_p;
   

	DEBUG("IMB_ioctl: cmd 0x%x arg 0x%x\n", cmd, arg);

    if (deviceExtensionP == NULL)
        return(-ENXIO);				//ENXIO: No such device or address

    if ((caddr_t)arg == NULL)
        return(-EINVAL);		//EINVAL: Invalid argument

	// First copy the smi data from user to kernel space
    if (copy_from_user((caddr_t)&smi, (caddr_t)arg, sizeof(smi)) == -1) {
		printk(KERN_ERR "imb_ioctl: copy_from_user failed\n");
        return(-EFAULT);
	}

    inputBufferLength  = smi.cbInBuffer;
    outputBufferLength = smi.cbOutBuffer; 


	DEBUG("imb_ioctl: inputBufferLength = %d outputBufferLength = %d\n",
			inputBufferLength, outputBufferLength);
    //
    // Case on the ioctl sent down by the user
    //
    switch ( cmd ) {

		case IOCTL_IMB_GET_ASYNC_MSG:
		//
		// check parameters
		//
		DEBUG("imb_ioctl: IOCTL_IMB_GET_ASYNC_MSG\n");
		if( inputBufferLength < sizeof( ImbAsyncRequest ) ) {
			// 
			// Too short an input buffer than we expected - invalid error
			//
			DEBUG("imb_ioctl: IOCTL_IMB_GET_ASYNC_MSG - InBufLen too short\n");
			return(-EINVAL);
		} else if ( outputBufferLength < MIN_ASYNC_RESP_SIZE )   {
			// 
			// Too short an output buffer than we expected - invalid error
			//
			DEBUG("imb_ioctl: IOCTL_IMB_GET_ASYNC_MSG - OutBufLen too short\n");
			return(-EINVAL);
		} else {
			//
			// get his request and try to get a queued message, If none are 
			// available and the timeout is non-zero, the request will sleep 
			// until timeout or message available
			//
			
			unsigned char AsyncReq[MAX_ASYNC_RESP_SIZE];
			unsigned char AsyncResp[MAX_ASYNC_RESP_SIZE];
			ImbAsyncRequest  *pAsyncReq;
			ImbAsyncResponse *pAsyncResp;

			pAsyncReq = (ImbAsyncRequest *)AsyncReq;
			pAsyncResp = (ImbAsyncResponse *)AsyncResp;

			//
			// Now we copy the Input Buffer data itself from user space.
			//
			if (copy_from_user( (void *)AsyncReq, (void *)smi.lpvInBuffer, 
						sizeof(ImbAsyncRequest)) == -1) {
				printk(KERN_ERR "imb_ioctl: copy_from_user of lpvInBuffer failed\n");
				return(-EFAULT);
			}

			length = outputBufferLength;

			status = ImbAsyncGetMessage( pAsyncReq, pAsyncResp,&length,
										&deviceExtensionP->asyncQ );

			//
			// Now copy response and length back into user-space
			//
			
			if (copy_to_user( 	(void *)smi.lpvOutBuffer,	
								(void *)pAsyncResp,  
								length
							) == -1) {
				printk(KERN_ERR "imb_ioctl: IOCTL_IMB_GET_ASYNC_MSG \
									copy_to_user to lpvOutBuffer failed\n");
				return(-EFAULT);
			}

			if ( copy_to_user(	(void *)smi.lpcbBytesReturned,
							  	(void *)&length,  
								sizeof(smi.lpcbBytesReturned)
							  ) == -1)  {
				printk(KERN_ERR "imb_ioctl: IOCTL_IMB_GET_ASYNC_MSG \
									 copy_to_user to lpvOutBuffer failed\n");
				return(-EFAULT);
			}
        }
        break;

    case IOCTL_IMB_SEND_MESSAGE:
        //
        // check parameters
        //
        if ( (inputBufferLength < MIN_IMB_REQ_BUF_SIZE) || 
        	 (outputBufferLength < MIN_IMB_RESP_BUF_SIZE) )   {
            //
            // Input or Output buffer is too small
            //
			DEBUG("imb_ioctl: IOCTL_IMB_SEND_MESSAGE - Buffer Len too short\n");
            return(-EINVAL);
        } else {
           //
           // let the command processing check the request contents
           //
			// probably bigger than needed check later 
			unsigned char imbReqBuffer[MAX_IMB_RESPONSE_SIZE + 8];
			unsigned char imbRespBuffer[MAX_IMB_RESPONSE_SIZE + 8];

    		ImbRequestBuffer   *pImbReq;
    		ImbResponseBuffer  *pImbResp;

			pImbReq = (ImbRequestBuffer *)imbReqBuffer;
			pImbResp = (ImbResponseBuffer *)imbRespBuffer;

			//
			// Now we copy the Input Buffer data itself from user space.
			//
			if (copy_from_user( (void *)pImbReq, (void *)smi.lpvInBuffer,  
						inputBufferLength) == -1) {
				printk(KERN_ERR "imb_ioctl: copy_from_user ImbRequestBuffer failed\n");
				return(-EFAULT);
			}

			//
    		// Setup the timer for this request and get the request semaphore
    		//
    		ImbOsStartReq( pImbReq->timeOut, deviceExtensionP );

    		if( (DWORD)(pImbReq->req.dataLength + MIN_IMB_REQ_BUF_SIZE) > 
							inputBufferLength ) {
				//
				//cleanup and return
				//
	    		ImbOsEndReq( deviceExtensionP );
				DEBUG("imb_ioctl: IOCTL_IMB_SEND_MESSAGE failed\n");
				return(-EINVAL);
			} else {
				length = outputBufferLength - MIN_IMB_RESP_BUF_SIZE;


				//
				// do the work of sending the request
				//
				status = SendImbRequest( 	pImbReq, 
											pImbResp, 
											&length,
											deviceExtensionP
										);


				if( status == IMB_SUCCESS ) {
					//
					//Now copy back the imb response and length to user space
					//
					if (copy_to_user( 	(void *)smi.lpvOutBuffer,
										(void *)pImbResp,  
										length
									) == -1) {
						printk(KERN_ERR
							"imb_ioctl: copy_to_user to lpvOutBuffer failed\n");
						return(-EFAULT);
					}

					if ( copy_to_user(	(void *)smi.lpcbBytesReturned,
										(void *)&length,  
										sizeof(smi.lpcbBytesReturned) 
									  ) == -1 ) {
						printk(KERN_ERR "imb_ioctl: copy_to_user to lpcbBytesReturned failed\n");
						return(-EFAULT);
					}
				} //IMB_SUCCESS

				//
				// All went well. Now clean up
				//
				DEBUG("imb_ioctl: IOCTL_IMB_SEND_MESSAGE done\n");
    			ImbOsEndReq( deviceExtensionP );
			} //else

			// need to do a ImbAsyncMessageProcess(deviceExtensionP); now???
        }
        break;

    case IOCTL_IMB_SHUTDOWN_CODE:
        {
			ShutdownCmdBuffer shutdownCmd;

			if (copy_from_user( (void *)&shutdownCmd, (void *)smi.lpvInBuffer, 
						sizeof(ShutdownCmdBuffer)) == -1) {
				printk(KERN_ERR "imb_ioctl: copy_from_user of shutdownCmd failed\n");
				return(-EFAULT);
			}

            //
            // check input parameters
            //
            if ( inputBufferLength < sizeof(ShutdownCmdBuffer) ) {
				DEBUG("imb_ioctl: IOCTL_IMB_SHUTDOWN_CODE InBufLen short\n");
                return(-EINVAL);
            }
            //
            // guarentee valid code
            //
            if (    shutdownCmd.code != SD_POWER_OFF   &&
                    shutdownCmd.code != SD_RESET       &&
                    shutdownCmd.code != SD_NO_ACTION       ) {

                deviceExtensionP->shutdownCode = SD_NO_ACTION;
                status = (long)INVALID_INPUT_BUFFER;
                break;
            }

            deviceExtensionP->shutdownCode = shutdownCmd.code;
            deviceExtensionP->delayTime    = shutdownCmd.delayTime;

			ImbShutdown(deviceExtensionP); // never return from here right?
            status = STATUS_SUCCESS;
        }

        break;
		
		//For DPC thread to register for anync messages.

		case IOCTL_IMB_REGISTER_ASYNC_OBJ:
			if (ImbUnixAsyncInit (deviceExtension) != 0) {
				int dummy;

				status = STATUS_SUCCESS;
				//
				//Now copy back the  a bogus imb response and length to 
				//user space since the imbapi expects it. We can use this
				//later to return a more meaningful value if necesssary.
				//
				dummy = 0xbabe;
				if (copy_to_user( 	(void *)smi.lpvOutBuffer,
									(void *)&dummy,  
									length
								) == -1) {
					printk(KERN_ERR
						"imb_ioctl: copy_to_user to lpvOutBuffer failed\n");
					return(-EFAULT);
				}
				
				dummy = sizeof(int); //sending back an int above.
				if ( copy_to_user(	(void *)smi.lpcbBytesReturned,
									(void *)&dummy,  
									sizeof(dummy)
								  ) == -1 ) {
					printk(KERN_ERR 
					"imb_ioctl: copy_to_user to lpcbBytesReturned failed\n");
					return(-EFAULT);
				}
			}
			else
				status = STATUS_UNSUCCESSFUL; //MR is this a good status?
		break;

		case IOCTL_IMB_DEREGISTER_ASYNC_OBJ:
			ImbUnixAsyncDeallocate(deviceExtension);
			status = STATUS_SUCCESS;
		break;

		case IOCTL_IMB_CHECK_EVENT:
			//app thread gets blocked for async events here till an
			// event is received and unblocked by ImbAsyncQueueMessage
			ImbUnixAsyncLock(deviceExtension);
		break;

		//This ioctl is commented out in imbapi.c 
		//case IOCTL_IMB_GET_DPC_MSG:

		// mmap does this for us, we don't use an ioctl for these. MR
		//    case IOCTL_IMB_MAP_MEMORY : 
		//    case IOCTL_IMB_UNMAP_MEMORY: 


		default:
			printk(KERN_ERR "imb_ioctl: ioctl cmd  0x%x not supported\n", cmd);
			return(-EINVAL);
    } // switch (cmd)

    return(status);
} //IMB_ioctl


/////////////////////////////////////////////////////////////
//  IMBmmap
/////////////////////////////////////////////////////////////
/*F*
//  Name:       IMBmmap
//  Purpose:    maps physical memory for user
//  Context:    called when user calls mmap() on device
//  Returns:    
//  Parameters: device number, offset (from phys addr 0, not from beginning
//              of device!), and protection info
//  Notes:      none
*F*/

static int
IMBmmap(struct file *filep, struct vm_area_struct *vma)
{
    off_t                   offset = vma->vm_offset;
    int                     ret;

	DEBUG("IMBmmap:vm_start 0x%x vm_end 0x%x\n",vma->vm_start,vma->vm_end);

    if (offset < 0) {
        printk(KERN_ERR "IMBmmap: called with invalid offset\n");
        return(-1);
    }

	if (remap_page_range(vma->vm_start, offset, 
							vma->vm_end - vma->vm_start, vma->vm_page_prot))
	{
		printk(KERN_ERR, "IMBmmap: mapping failed\n");
		return(-EAGAIN);
	}
	
	DEBUG("IMBmmap: mmap success\n");

	//vma->vm_inode = what_goes_here;

    return(0);
} //IMBmmap


/////////////////////////////////////////////////////////////////////
// ImbShutdown
/////////////////////////////////////////////////////////////////////
/*F*
//  Name:       ImbShutdown
//  Purpose:    This is the dispatch routine for processing shutdown request.
//  Context:    Called by operating system during system exit.
//  Returns:    STATUS_SUCCESS
//  Parameters:
//              deviceExtension    Pointer to driver object created by system.
//  Notes:      none
*F*/
static NTSTATUS
ImbShutdown ( PDEVICE_EXTENSION deviceExtension)
{
    NTSTATUS            status;

    switch ( deviceExtension->shutdownCode ) {

    case SD_POWER_OFF:
    case SD_RESET   :
        //
        //
        // Get the request semaphore, to  avoid colliding with the
        // main request thread. Also starts the request timer
        //
        ImbOsStartReq( DEFAULT_MESSAGE_TIMEOUT, deviceExtension );
        //
        // execute the action
        //
        ImbShutdownSystem(  deviceExtension->shutdownCode,
                            deviceExtension->delayTime,
                            deviceExtension
                         );
        //
        // release the request semaphore
        //
        ImbOsEndReq( deviceExtension );

        status = STATUS_SUCCESS;
        break;

    case SD_NO_ACTION:

        status = STATUS_SUCCESS;
        break;

    default:
        status = INVALID_INPUT_BUFFER ;
        break;
    }
    return status;
} //ImbShutdown

