/*M*
//                    INTEL CORPORATION PROPRIETARY INFORMATION
//      This software is supplied under the terms of a license agreement or
//      nondisclosure agreement with Intel Corporation and may not be copied
//      or disclosed except in accordance with the terms of that agreement.
//            Copyright (c) 1995 Intel Corporation.  All Rights Reserved.
//
//  PVCS:
//      $Workfile:   imbapi.c  $
//      $Revision:   1.5  $
//      $Modtime:   24 Apr 2000 17:55:48  $
//      $Author:   mramacha  $
//
//  Purpose:    This file contains the entry point that opens the IMB device in
//              order to issue the  IMB driver API related IOCTLs.
//              This file implements the IMB driver API for the Server 
//				Management Agents
//
//  
*M*/

/*
 * $Log:   //cospublic004/shareddata/isc/ISC2x/LNWORK/SRC/imbapi/imbapi.c.v  $
 * 
 *    Rev 1.5   24 Apr 2000 16:56:42   mramacha
 * Changed the include file from windows.h to windefs.h
 * 
 * 
 *    Rev 1.4   23 Feb 2000 10:12:14   mramacha
 * Debug statement was not in debug mode and
 * will cause confusing error messages. Changed
 * to display message only when debugging is turned
 * on.
 * 
 *    Rev 1.3   21 Oct 1999 18:50:10   mramacha
 * mmap call had its arguments reversed
 * and prevented us from accessing the bios area.
 * 
 *    Rev 1.2   14 Oct 1999 13:05:12   mramacha
 * Changed from /dev/mem to /dev/imb for
 * mem mapping.
 * 
 *    Rev 1.1   09 Sep 1999 12:00:36   mramacha
 * Linux updates. 
 * 
 *    Rev 1.0   05 Sep 1999 17:20:30   mramacha
 * Linux checkin
 * 
 * 
 */

/*
 *
 *	Note: This file is derived from the NTWORK version of the imbapi.c
 *	It was decided to create OS specific ones for Linux and Solaris. 
 *	The following log has been preserved to show that it has all the fixes
 *	that went into the imbapi.c upto Rev 1.12 in the 2.2 NTWORK branch.
 *
 *    Rev 1.12   02 Sep 1999 11:03:16   jpaiasno
 * Fixed logic problem related to copying 
 * i2c response
 * 
 *    Rev 1.11   19 Aug 1999 11:38:20   sganeshx
 * 1. modified function 'UnRegisterForImbAsyncMessageNotification'
 *     to accept second parameter.
 * 
 *    Rev 1.10   27 Jul 1999 09:52:34   jpaiasno
 * Changed the size of BMC  communication buffer
 * to 64. This change was done to support all
 * IPMI 1.0 based commands.
 * 
 * 
 *    Rev 1.9   Jun 14 1999 11:07:22   sganeshx
 * 1. added  check for BMC in 'SendTimedImbpRequest'
 *     
 * 
 *    Rev 1.8   May 26 1999 18:30:12   sganeshx
 * 1. eliminate syntax error due to missing
 *     semicolon (only in Unix section).
 * 
 *    Rev 1.7   May 26 1999 17:22:56   sganeshx
 * 1. modified function 'open_mismic' to fix
 *     compilation errors because local variables
 *     were not defined
 * 
 *    Rev 1.6   May 07 1999 12:04:34   sganeshx
 * 1. added new constant for response length
 *     from Get Device ID command (IPMI v1.0)
 *     IPMI10_GET_DEVICE_ID_RESP_LENGTH
 * 
 * 2. use (IPMI10_GET_DEVICE_ID_RESP_LENGTH-1)
 *     to check for IPMI version.
 * 
 *    Rev 1.5   04 May 1999 15:47:52   jpaiasno
 * Added IPMI 1.0 code to 'ImbGetAsyncMessage'
 * 
 *    Rev 1.4   28 Apr 1999 17:09:46   jpaiasno
 * Fixed the problem for NT as well as UnixWare
 * (response  length check to be >1 to account for completion code)
 * 
 *    Rev 1.3   28 Apr 1999 17:00:26   jpaiasno
 * fixed response data length check logic
 * in 'SendDataTimedImbp' function
 * 
 *    Rev 1.2   27 Apr 1999 14:15:16   jpaiasno
 * Added code to support IPMI 1.0 for IPMB
 * requests.
 * 
 *    Rev 1.1   08 Dec 1998 11:15:40   jpaiasno
 * Added code to support DPC related
 * Async. events from the EMP console
 */




#define IMB_API

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include "../ntioinc/windefs.h"
#include "../ntioinc/smi.h"
#include "../imb/imb_if.h"
#include "../imb/imb_log.h"

#include "imbapi.h"

#define PAGESIZE EXEC_PAGESIZE

#define IMB_DEVICE "/dev/imb"

//Just to make the DEBUG code cleaner.

#ifdef LINUX_DEBUG
#define DEBUG(format, args...) printf(format, ##args)
#else
#define DEBUG(format, args...)
#endif

// uncomment out the #define below or use -DLINUX_DEBUG_MAX in the makefile 
// if you want a dump of the memory to debug mmap system call in
// MapPhysicalMemory() below.
// 
//#define LINUX_DEBUG_MAX


//keep it simple. use global varibles for event objects and handles
//pai 10/8

// UnixWare should eventually have its own source code file. Right now
// new code has been added based on the exsisting policy of using
// pre-processor directives to separate os-specific code (pai 11/21)

HANDLE  AsyncEventHandle = 0;
static void *  AsyncEventObject = 0; 
static int  IpmiVersion;

/////////////////////////////////////////////////////////////////////////////////////////
//  GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////////////////

IO_STATUS_BLOCK NTstatus; //dummy place holder. See deviceiocontrol.
static HANDLE   hDevice1;
//mutex_t deviceMutex;

static HANDLE hDevice;

//////////////////////////////////////////////////////////////////////
// open_imb
//////////////////////////////////////////////////////////////////////
/*F*
//  Name:       open_imb
//  Purpose:    To open imb device
//  Context:    Called from each routine to make sure that open is done.
//  Returns:    returns 0 for Fail and 1 for Success, sets hDevice to open
//              handle.
//  Parameters: none
//  Notes:      none
*F*/

int open_imb()
{
/* This routine will be called from all other routineds before doing any
   interfacing with imb driver. It will open only once. */
	IMBPREQUESTDATA                requestData;
	BYTE						   respBuffer[16];
	DWORD						   respLength;
	BYTE						   completionCode;

	int my_ret_code;

  if(hDevice1 == 0)
  {
			DEBUG("%s: opening the driver\n", __FUNCTION__);

			//O_NDELAY flag will cause problems later when driver makes
			//you wait. Hence removing it.
		    //if ((hDevice1 = open(IMB_DEVICE,O_RDWR|O_NDELAY)) <0) 
		    if ((hDevice1 = open(IMB_DEVICE,O_RDWR)) <0) 
			{
				char buf[128];

				hDevice1  = 0;
				sprintf(buf,"%s %s: open(%s) failed\n", 
							__FILE__,__FUNCTION__,IMB_DEVICE);
				perror(buf);
				return (0);
			}	
			
			//detect the IPMI version for processing requests later.
			// This is a crude but most reliable method to differentiate
			// between old IPMI versions and the 1.0 version. If we had used the
			// version field instead then we would have had to revalidate all the
			// older platforms (pai 4/27/99)
		
			requestData.cmdType            = GET_DEVICE_ID;
			requestData.rsSa               = BMC_SA;
			requestData.rsLun              = BMC_LUN;
			requestData.netFn              = APP_NETFN ;
			requestData.busType            = PUBLIC_BUS;
			requestData.data               = NULL; 
			requestData.dataLength			= 0;
			respLength					    = 16;

			DEBUG("%s: opened driver, getting IPMI version\n", __FUNCTION__);

			if (
				 (( my_ret_code = SendTimedImbpRequest (
						&requestData,                
						(DWORD)400,
						 respBuffer, 
					     (int *)&respLength,    
						  &completionCode  
						))  != ACCESN_OK ) || ( completionCode != 0)
			)
			{
				printf("%s: SendTimedImbpRequest error. Ret = %d CC = 0x%X\n",
					__FUNCTION__, my_ret_code, completionCode);
					hDevice1 = 0;
					return (0);
			}

			if (respLength < (IPMI10_GET_DEVICE_ID_RESP_LENGTH-1))
				IpmiVersion = IPMI_09_VERSION;
			else
				IpmiVersion = IPMI_10_VERSION;


			DEBUG("%s: IPMI version 0x%x\n", __FUNCTION__, IpmiVersion);	
		
/*
//initialise a mutex
		if(mutex_init(&deviceMutex , USYNC_THREAD, NULL) != 0)
		{
			return(0);
		}

*/

	}
    

	return (1);
}


/////////////////////////////////////////////////////////////////////////////
// DeviceIoControl 
/////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       DeviceIoControl 
//  Purpose:    Simulate NT DeviceIoControl using unix calls and structures.
//  Context:    called for every NT DeviceIoControl
//  Returns:    FALSE for fail and TRUE for success. Same as standarad NTOS call
//              as it also sets Ntstatus.status.
//  Parameters: Standard NT call parameters, see below.
//  Notes:      none
*F*/
static BOOL
DeviceIoControl(
	HANDLE 			dummey_hDevice, /* handle of device */
	DWORD 			dwIoControlCode, /* control code of operation to perform*/
	LPVOID 			lpvInBuffer, /* address of buffer for input data */
	DWORD 			cbInBuffer, /* size of input buffer */
	LPVOID 			lpvOutBuffer, /* address of output buffer */
	DWORD 			cbOutBuffer, /* size of output buffer */
	LPDWORD 		lpcbBytesReturned, /* address of actual bytes of output */
	LPOVERLAPPED 	lpoOverlapped /* address of overlapped struct */
	)
{
	struct smi s;
	int rc;
	int ioctl_status;

  	rc = open_imb();
  	if (rc == 0) {
    	return FALSE;
    }

	/*
		//lock the mutex, before making the request....
		if(mutex_lock(&deviceMutex) != 0)
		{
			return(FALSE);
		}
	*/
	DEBUG("%s: ioctl cmd = 0x%x ", __FUNCTION__,dwIoControlCode);
	DEBUG("cbInBuffer %d cbOutBuffer %d\n", cbInBuffer, cbOutBuffer);

  	s.lpvInBuffer = lpvInBuffer;
  	s.cbInBuffer = cbInBuffer;
  	s.lpvOutBuffer = lpvOutBuffer;
  	s.cbOutBuffer = cbOutBuffer;
  	s.lpcbBytesReturned = lpcbBytesReturned;
  	s.lpoOverlapped = lpoOverlapped;
  	s.ntstatus = (LPVOID)&NTstatus; //dummy place holder. Linux IMB driver
									//doesn't return status or info via it.

  	if ( (ioctl_status = ioctl(hDevice1, dwIoControlCode,&s) ) <0) {
 	  	DEBUG("%s %s: ioctl cmd = 0x%x failed\n", 
			__FILE__,__FUNCTION__,dwIoControlCode);
		//      mutex_unlock(&deviceMutex);
    	return FALSE;
    }
	//      mutex_unlock(&deviceMutex);

	DEBUG("%s: ioctl_status %d  bytes returned =  %d \n",
		 __FUNCTION__, ioctl_status,  *lpcbBytesReturned);

//MR commented this just as in Sol1.10. lpcbBytesReturned has the right data
//  	*lpcbBytesReturned = NTstatus.Information;
  
	if (ioctl_status == STATUS_SUCCESS) {
		DEBUG("%s returning true\n", __FUNCTION__);
     	return (TRUE);
	}
	else {
		DEBUG("%s returning false\n", __FUNCTION__);
     	return (FALSE);
     }
}

//Used only by UW. Left here for now. IMB driver will not accept this
//ioctl.
ACCESN_STATUS
StartAsyncMesgPoll()
{

	DWORD   retLength;
	BOOL    status;
   
	status = DeviceIoControl (      hDevice,
								IOCTL_IMB_POLL_ASYNC,
								NULL,
								0,
								NULL,
								0,
								& retLength,
								0
							);
 
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( status == TRUE ) {
		return ACCESN_OK;
	} else {
		return ACCESN_ERROR;
	}

}

/////////////////////////////////////////////////////////////////////////////////////////////////
// SendTimedI2cRequest
/////////////////////////////////////////////////////////////////////////////////////////////////

/*F*
//  Name:       SendTimedI2cRequest 
//  Purpose:    This function sends a request to a I2C device
//  Context:    Used by Upper level agents (sis modules) to access dumb I2c devices 
//  Returns:    ACCESN_OK  else error status code
//  Parameters: 
//     reqPtr
//     timeOut
//     respDataPtr
//     respLen
//  Notes:      none
*F*/

ACCESN_STATUS
SendTimedI2cRequest (
		I2CREQUESTDATA 		*reqPtr,         // I2C request
		int                 timeOut,         // how long to wait, mSec units
		BYTE 				*respDataPtr,    // where to put response data
		int 				*respDataLen,    // size of response buffer and
											 // size of returned data
		BYTE 				*completionCode  // request status from BMC
	)
{
	BOOL  					status;
    BYTE                  	responseData[MAX_IMB_RESP_SIZE];
	ImbResponseBuffer *     resp = (ImbResponseBuffer *) responseData;
	DWORD                   respLength = sizeof( responseData );
    BYTE                    requestData[MAX_IMB_RESP_SIZE];
	ImbRequestBuffer *      req = (ImbRequestBuffer *) requestData;

	struct WriteReadI2C {   // format of a write/read I2C request
		BYTE    busType;
		BYTE    rsSa;
		BYTE    count;
		BYTE    data[1];
	} * wrReq = (struct WriteReadI2C *) req->req.data;

#define MIN_WRI2C_SIZE  3       // size of write/read request minus any data


	//
	// If the Imb driver is not present return AccessFailed
	//

	req->req.rsSa           = BMC_SA;
	req->req.cmd            = WRITE_READ_I2C;
	req->req.netFn          = APP_NETFN;
	req->req.rsLun          = BMC_LUN;
	req->req.dataLength     = reqPtr->dataLength + MIN_WRI2C_SIZE;

	wrReq->busType          = reqPtr->busType;
	wrReq->rsSa                     = reqPtr->rsSa;
	wrReq->count            = reqPtr->numberOfBytesToRead;
	
	memcpy( wrReq->data, reqPtr->data, reqPtr->dataLength );

	req->flags      = 0;
	req->timeOut    = timeOut * 1000;       // convert to uSec units

	status = DeviceIoControl(       hDevice,
								IOCTL_IMB_SEND_MESSAGE,
								requestData,
								sizeof( requestData ),
								& responseData,
								sizeof( responseData ),
								& respLength,
								NULL
							  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( status != TRUE ) {
		DWORD error = GetLastError();
		return ACCESN_ERROR;
	}
	if( respLength == 0 ) {
		return ACCESN_ERROR;
	}

	//
	// give the caller his response
	//
	*completionCode = resp->cCode;
	*respDataLen    = respLength - 1;

	if(( *respDataLen ) && (respDataPtr))
		memcpy( respDataPtr, resp->data, *respDataLen);

	return ACCESN_OK;

}

//This is not a  API exported by the driver in stricter sense. It is 
//added to support EMP functionality. Upper level software could have 
//implemented this function.(pai 5/4/99)
/////////////////////////////////////////////////////////////////////////////
// SendTimedEmpMessageResponse
/////////////////////////////////////////////////////////////////////////////

/*F*
//  Name:       SendTimedEmpMessageResponse 
//  Purpose:    This function sends a response message to the EMP port
//  Context:     
//  Returns:    OK  else error status code
//  Parameters: 
//  
//  Notes:      none
*F*/

ACCESN_STATUS
SendTimedEmpMessageResponse (
		ImbPacket *ptr,       // pointer to the original request from EMP
		char      *responseDataBuf,
		int       responseDataLen,
		int       timeOut         // how long to wait, in mSec units
	)
{
	BOOL                    status;
    BYTE                    responseData[MAX_IMB_RESP_SIZE];
	ImbResponseBuffer *     resp = (ImbResponseBuffer *) responseData;
	DWORD                   respLength = sizeof( responseData );
    BYTE                    requestData[MAX_IMB_RESP_SIZE];
	ImbRequestBuffer *      req = (ImbRequestBuffer *) requestData;
	int 					i,j;

	//form the response packet first
	req->req.rsSa           =  BMC_SA;
	if (IpmiVersion ==	IPMI_09_VERSION)
	req->req.cmd            =  WRITE_EMP_BUFFER;
	else 
	req->req.cmd            =  SEND_MESSAGE;	
	req->req.netFn          =  APP_NETFN;
	req->req.rsLun          =  0;

	i = 0;
	if (IpmiVersion !=	IPMI_09_VERSION)
		req->req.data[i++] 	= EMP_CHANNEL;
		
	req->req.data[i++]    =  ptr->rqSa;
	req->req.data[i++]      =  (((ptr->nfLn & 0xfc) | 0x4) | ((ptr->seqLn) & 0x3));
	if (IpmiVersion ==	IPMI_09_VERSION)
		req->req.data[i++]    = ((~(req->req.data[0] +  req->req.data[1])) +1);
	else
		req->req.data[i++]    = ((~(req->req.data[1] +  req->req.data[2])) +1);

	req->req.data[i++]      =  BMC_SA; //though software is responding, we have to
								   //provide BMC's slave address as responder
								   //address.
	
	req->req.data[i++]      = ( (ptr->seqLn & 0xfc) | (ptr->nfLn & 0x3) );

	req->req.data[i++]      = ptr->cmd;
	for ( j = 0 ; j < responseDataLen ; ++j,++i)
	   req->req.data[i] = responseDataBuf[j];

	 req->req.data[i] = 0;
	 if (IpmiVersion ==	IPMI_09_VERSION)
		 j = 0;
	 else 
		 j = 1;
	for ( ; j < ( i -3); ++j)
		 req->req.data[i] += req->req.data[j+3];
	req->req.data[i]  = ~(req->req.data[i]) +1;
	++i;
	req->req.dataLength     = i;

	req->flags      = 0;
	req->timeOut    = timeOut * 1000;       // convert to uSec units


	status = DeviceIoControl(       hDevice,
								IOCTL_IMB_SEND_MESSAGE,
								requestData,
								sizeof(requestData),
								responseData,
								sizeof( responseData ),
								& respLength,
								NULL
							  );

	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);


	if ( (status != TRUE) || (respLength != 1) || (responseData[0] != 0) )
	{
		return ACCESN_ERROR;
	}
	return ACCESN_OK;
}

/////////////////////////////////////////////////////////////////////////////
// SendTimedImbpRequest
/////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       SendTimedImbpRequest 
//  Purpose:    This function sends a request for BMC implemented function
//  Context:    Used by Upper level agents (sis modules) to access BMC implemented functionality. 
//  Returns:    OK  else error status code
//  Parameters: 
//     reqPtr
//     timeOut
//     respDataPtr
//     respLen
//  Notes:      none
*F*/
ACCESN_STATUS
SendTimedImbpRequest (
		IMBPREQUESTDATA *reqPtr,         // request info and data
		int             timeOut,         // how long to wait, in mSec units
		BYTE 			*respDataPtr,    // where to put response data
		int 			*respDataLen,    // how much response data there is
		BYTE 			*completionCode  // request status from dest controller
	)
{
	BOOL                    status;
    BYTE                    responseData[MAX_BUFFER_SIZE];
	ImbResponseBuffer *     resp = (ImbResponseBuffer *) responseData;
	DWORD                   respLength = sizeof( responseData );
    BYTE                    requestData[MAX_BUFFER_SIZE];
	ImbRequestBuffer *      req = (ImbRequestBuffer *) requestData;


	req->req.rsSa           = reqPtr->rsSa;
	req->req.cmd            = reqPtr->cmdType;
	req->req.netFn          = reqPtr->netFn;
	req->req.rsLun          = reqPtr->rsLun;
	req->req.dataLength     = reqPtr->dataLength;

	memcpy( req->req.data, reqPtr->data, reqPtr->dataLength );

	req->flags              = 0;
	req->timeOut    = timeOut * 1000;       // convert to uSec units

	DEBUG("%s: rsSa 0x%x cmd 0x%x netFn 0x%x rsLun 0x%x\n", __FUNCTION__,
			req->req.rsSa, req->req.cmd, req->req.netFn, req->req.rsLun);


	status = DeviceIoControl(       hDevice,
								IOCTL_IMB_SEND_MESSAGE,
								requestData,
								sizeof( requestData ),
								& responseData,
								sizeof( responseData ),
								& respLength,
								NULL
							  );
	DEBUG("%s: DeviceIoControl returned status = %d\n",__FUNCTION__, status);

	if( status != TRUE ) {
		DWORD error = GetLastError();
		return ACCESN_ERROR;
	}
	if( respLength == 0 ) {
		return ACCESN_ERROR;
	}
	//
	// give the caller his response
	//
  


	*completionCode = resp->cCode;
	*respDataLen    = 0;

    if(( respLength > 1 ) && ( respDataPtr))
	{
		*respDataLen    = respLength - 1;
		memcpy( respDataPtr, resp->data, *respDataLen);
	}


	return ACCESN_OK;
}


///////////////////////////////////////////////////////////////////////////
//SendAsyncImbpRequest
///////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       SendAsyncImbpRequest
//  Purpose:    This function sends a request for Asynchronous IMB implemented function
//  Context:    Used by Upper level agents (sis modules) to access Asynchronous IMB implemented functionality. 
//  Returns:    OK  else error status code
//  Parameters: 
//     reqPtr  Pointer to Async IMB request
//     seqNo   Sequence Munber
//  Notes:      none
*F*/
ACCESN_STATUS
SendAsyncImbpRequest (
		IMBPREQUESTDATA *reqPtr,  // request info and data
		BYTE *          seqNo     // sequence number used in creating IMB msg
	)
{

	BOOL                    status;
    BYTE                    responseData[MAX_IMB_RESP_SIZE];
	ImbResponseBuffer *     resp = (ImbResponseBuffer *) responseData;
	DWORD                   respLength = sizeof( responseData );
    BYTE                    requestData[MAX_IMB_RESP_SIZE];
	ImbRequestBuffer *      req = (ImbRequestBuffer *) requestData;

	req->req.rsSa           = reqPtr->rsSa;
	req->req.cmd            = reqPtr->cmdType;
	req->req.netFn          = reqPtr->netFn;
	req->req.rsLun          = reqPtr->rsLun;
	req->req.dataLength     = reqPtr->dataLength;

	memcpy( req->req.data, reqPtr->data, reqPtr->dataLength );

	req->flags              = NO_RESPONSE_EXPECTED;
	req->timeOut    = 0;    // no timeouts for async sends

	status = DeviceIoControl(       hDevice,
								IOCTL_IMB_SEND_MESSAGE,
								requestData,
								sizeof( requestData ),
								& responseData,
								sizeof( responseData ),
								& respLength,
								NULL
							  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( status != TRUE ) {
		DWORD error = GetLastError();
		return ACCESN_ERROR;
	}
	if( respLength != 2 ) {
		return ACCESN_ERROR;
	}
	//
	// give the caller his sequence number
	//
	*seqNo = resp->data[0];

	return ACCESN_OK;

}

/////////////////////////////////////////////////////////////////////////////
//GetAsyncImbpMessage
/////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       GetAsyncImbpMessage
//  Purpose:    This function gets the next available async message with a message id
//                              greater than SeqNo. The message looks like an IMB packet
//                              and the length and Sequence number is returned
//  Context:    Used by Upper level agents (sis modules) to access Asynchronous IMB implemented functionality. 
//  Returns:    OK  else error status code
//  Parameters: 
//     msgPtr  Pointer to Async IMB request
//     msgLen  Length 
//     timeOut Time to wait 
//     seqNo   Sequence Munber
//  Notes:      none
*F*/

ACCESN_STATUS
GetAsyncImbpMessage (
		ImbPacket *     msgPtr,         // request info and data
		DWORD 			*msgLen,        // IN - length of buffer, OUT - msg len
		DWORD			timeOut,        // how long to wait for the message
		ImbAsyncSeq 	*seqNo,         // previously returned seq number
										// (or ASYNC_SEQ_START)
		DWORD			channelNumber
	)
{

	BOOL                   status;
    BYTE                   responseData[MAX_ASYNC_RESP_SIZE], lun;
	ImbAsyncResponse *     resp = (ImbAsyncResponse *) responseData;
	DWORD                  respLength = sizeof( responseData );
	ImbAsyncRequest        req;

	while(1)
	{


		if( (msgPtr == NULL) || (msgLen == NULL) || ( seqNo == NULL) )
				return ACCESN_ERROR;

				req.timeOut   = timeOut * 1000;       // convert to uSec units
				req.lastSeq   = *seqNo;


			status = DeviceIoControl(       hDevice,
								IOCTL_IMB_GET_ASYNC_MSG,
								& req,
								sizeof( req ),
								& responseData,
								sizeof( responseData ),
								& respLength,
								NULL
							  );

	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

			if( status != TRUE ) {
				DWORD error = GetLastError();
				//
				// handle "msg not available" specially. it's different from
				// a random old error.
				//
				switch( error ) {
					case IMB_MSG_NOT_AVAILABLE:
												return ACCESN_END_OF_DATA;
					default:
											return ACCESN_ERROR;
				}
				return ACCESN_ERROR;
			}
			if( respLength < MIN_ASYNC_RESP_SIZE ) {
					return ACCESN_ERROR;
			}
			respLength -= MIN_ASYNC_RESP_SIZE;

			if( *msgLen < respLength ) {
					return ACCESN_ERROR;
			}


			//same code as in NT section
			if ( IpmiVersion == IPMI_09_VERSION)
			{

				switch( channelNumber) {
					case IPMB_CHANNEL:
								lun = IPMB_LUN;
								 break;

					case  EMP_CHANNEL:
								lun = EMP_LUN;
							  	break;

					default:
								lun = RESERVED_LUN;
								break;
				}

				if ( (lun == RESERVED_LUN) || 
					 (lun !=  ((((ImbPacket *)(resp->data))->nfLn) & 0x3 )) 
					)
				{
						*seqNo = resp->thisSeq;
						continue;
				}


				memcpy( msgPtr, resp->data, respLength );
				*msgLen = respLength;
				
			}	
			else 
			{
				//it's a 1.0 or  above version 	

				if (resp->data[0] != (BYTE)channelNumber)
				{
					*seqNo = resp->thisSeq;
					continue;
				}

				memcpy( msgPtr, &(resp->data[1]), respLength-1 );
				*msgLen = respLength-1;
						

			}
	
		//
		// give the caller his sequence number
		//
		*seqNo = resp->thisSeq;

		return ACCESN_OK;

	} //while (1) 
}

  



/////////////////////////////////////////////////////////////////////////////////////////
//IsAsyncMessageAvailable
/////////////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       IsMessageAvailable
//  Purpose:    This function waits for an Async Message  
//
//  Context:    Used by Upper level agents access Asynchronous IMB based
//              messages   
//  Returns:    OK  else error status code
//  Parameters: 
//               eventId
//    
//  Notes:     This call will block the calling thread if no Async events are 
//                              are available in the queue.
//
*F*/
ACCESN_STATUS
IsAsyncMessageAvailable (unsigned int   eventId )
{
    int 	dummy;
    int 	respLength = 0;
    BOOL  	status;

 // confirm that app is not using a bad Id


	if (  AsyncEventHandle  != (HANDLE) eventId)
	  	return ACCESN_ERROR;

	status = DeviceIoControl(hDevice,
				   IOCTL_IMB_CHECK_EVENT,
				   &AsyncEventHandle,
				    sizeof(HANDLE ),
				    &dummy,
					sizeof(int),
					(LPDWORD) & respLength,
					NULL
				  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( status != TRUE )
		return  ACCESN_ERROR;

	
	return ACCESN_OK;
}


//I have retained this commented code because later we may want to use 
//DPC message specific Processing (pai 11/21)

#ifdef NOT_COMPILED_BUT_LEFT_HERE_FOR_NOW

/////////////////////////////////////////////////////////////////////////////////////////
//GetAsyncDpcMessage
/////////////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       GetAsyncDpcMessage
//  Purpose:    This function gets the next available async message from
//                              the DPC client. 
//
//  Context:    Used by Upper level agents access Asynchronous IMB based
//              messages sent by the DPC client.  
//  Returns:    OK  else error status code
//  Parameters: 
//     msgPtr  Pointer to Async IMB request
//         msgLen  Length 
//         timeOut Time to wait 
//     seqNo   Sequence Munber
//  Notes:     This call will block the calling thread if no Async events are 
//                              are available in the queue.
//
*F*/

ACCESN_STATUS
GetAsyncDpcMessage (
		ImbPacket *             msgPtr,         // request info and data
		DWORD *                 msgLen,         // IN - length of buffer, OUT - msg len
		DWORD                   timeOut,        // how long to wait for the message
		ImbAsyncSeq *   seqNo,          // previously returned seq number (or ASYNC_SEQ_START)
	)
{
	BOOL                            status;
    BYTE                                responseData[MAX_ASYNC_RESP_SIZE];
	ImbAsyncResponse *      resp = (ImbAsyncResponse *) responseData;
	DWORD                           respLength = sizeof( responseData );
	ImbAsyncRequest         req;

	if( msgPtr == NULL || msgLen == NULL || seqNo == NULL )
		return ACCESN_ERROR;

	req.lastSeq             = *seqNo;


	hEvt = CreateEvent (NULL, TRUE, FALSE, NULL) ;
	if (!hEvt) {
		return ACCESN_ERROR;
	}

	status = DeviceIoControl(       hDevice,
								IOCTL_IMB_GET_DPC_MSG,
								& req,
								sizeof( req ),
								& responseData,
								sizeof( responseData ),
								& respLength,
								&ovl
							  );

	if( status != TRUE ) {
		DWORD error = GetLastError();
		//
		// handle "msg not available" specially. it's different from
		// a random old error.
		//
		//
	if (!status)
	{
			switch (error )
				{
						case ERROR_IO_PENDING:

								WaitForSingleObject (hEvt, INFINITE) ;
								ResetEvent (hEvt) ;
								break;

						case IMB_MSG_NOT_AVAILABLE:

							    CloseHandle(hEvt);
								return ACCESN_END_OF_DATA;

						default:
								CloseHandle(hEvt);
								return ACCESN_ERROR;
								
			}
	}



		if ( 
		( GetOverlappedResult(hDevice,   
							&ovl,    
							(LPDWORD)&respLength, 
							TRUE 
					) == 0 ) || (respLength <= 0)
		)

		{

			CloseHandle(hEvt);
			return ACCESN_ERROR;

		}


	}
	
	if( respLength < MIN_ASYNC_RESP_SIZE ) {
		CloseHandle(hEvt);
		return ACCESN_ERROR;
	}

	respLength -= MIN_ASYNC_RESP_SIZE;

	if( *msgLen < respLength ) {

		// The following code should have been just return ACCESN_out_of_range
		CloseHandle(hEvt);
		return ACCESN_ERROR;
	}

	memcpy( msgPtr, resp->data, respLength );

	*msgLen = respLength;
	//
	// give the caller his sequence number
	//
	*seqNo = resp->thisSeq;

	CloseHandle(hEvt);


	return ACCESN_OK;

}
#endif //NOT_COMPILED_BUT_LEFT_HERE_FOR_NOW



/////////////////////////////////////////////////////////////////////////////////////////
//RegisterForImbAsyncMessageNotification
/////////////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       RegisterForImbAsyncMessageNotification
//  Purpose:    This function Registers the calling application  
//                              for Asynchronous notification when a sms message
//                              is available with the IMB driver.                       
//
//  Context:    Used by Upper level agents to know that an async
//                              SMS message is available with the driver.  
//  Returns:    OK  else error status code
//  Parameters: 
//    handleId  pointer to the registration handle
//
//  Notes:      The calling application should use the returned handle to 
//              get the Async messages..
*F*/
ACCESN_STATUS
RegisterForImbAsyncMessageNotification (unsigned int *handleId)

{
	BOOL      status;
	DWORD     respLength ;
	int       dummy;

	//allow  only one app to register 

	if( (handleId  == NULL ) || (AsyncEventHandle) )
		return ACCESN_ERROR;


	status = DeviceIoControl(hDevice,
							IOCTL_IMB_REGISTER_ASYNC_OBJ,
							&dummy,
							sizeof( int ),
							&AsyncEventHandle,
							(DWORD)sizeof(HANDLE ),
							(LPDWORD) & respLength,
							NULL
						  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( (respLength != sizeof(int))  || (status != TRUE ))
		return  ACCESN_ERROR;

	*handleId = (unsigned int) AsyncEventHandle;
	
	DEBUG("handleId = %x AsyncEventHandle %x\n", *handleId, AsyncEventHandle);
	return ACCESN_OK;
}





/////////////////////////////////////////////////////////////////////////////////////////
//UnRegisterForImbAsyncMessageNotification
/////////////////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       UnRegisterForImbAsyncMessageNotification
//  Purpose:    This function un-registers the calling application  
//                              for Asynchronous notification when a sms message
//                              is available with the IMB driver.                       
//
//  Context:    Used by Upper level agents to un-register
//                              for  async. notification of sms messages  
//  Returns:    OK  else error status code
//  Parameters: 
//    handleId  pointer to the registration handle
//	  iFlag		value used to determine where this function was called from
//				_it is used currently on in NetWare environment_
//
//  Notes:      
*F*/
ACCESN_STATUS
UnRegisterForImbAsyncMessageNotification (unsigned int handleId, int iFlag)

{
	BOOL		status;
	DWORD		respLength ;
	int			dummy;

	iFlag = iFlag;	/* to keep compiler happy  We are not using this flag*/

	if (  AsyncEventHandle  != (HANDLE) handleId)
	  return ACCESN_ERROR;

	status = DeviceIoControl(hDevice,
							IOCTL_IMB_DEREGISTER_ASYNC_OBJ,
							&AsyncEventHandle,
							(DWORD)sizeof(HANDLE ),
							&dummy,
							(DWORD)sizeof(int ),
							(LPDWORD) & respLength,
							NULL
						  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if( status != TRUE )
		return  ACCESN_ERROR;

	return ACCESN_OK;
}


/////////////////////////////////////////////////////////////////////////////
// SetShutDownCode
/////////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       SetShutDownCode
//  Purpose:    To set the shutdown  action code
//  Context:    Called by the System Control Subsystem
//  Returns:    none
//  Parameters: 
//              code  shutdown action code which can be either
//              SD_NO_ACTION, SD_RESET, SD_POWER_OFF as defined in imb_if.h
*F*/
		
ACCESN_STATUS 
SetShutDownCode (
		int 	delayTime, 			 // time to delay in 100ms units
		int 	code                // what to do when time expires
	)
{   
	DWORD					retLength;
	BOOL                    status;
	ShutdownCmdBuffer       cmd;

	//
	// If Imb driver is not present return AccessFailed
	//
	if(hDevice == INVALID_HANDLE_VALUE)
		return ACCESN_ERROR;

	cmd.code        = code;
	cmd.delayTime   = delayTime;

	status = DeviceIoControl( hDevice,
							IOCTL_IMB_SHUTDOWN_CODE,
							& cmd,
							sizeof( cmd ),
							NULL,
							0,
							& retLength,
							NULL
						  );
	DEBUG("%s: DeviceIoControl status = %d\n",__FUNCTION__, status);

	if(status == TRUE)
		return ACCESN_OK;
	else
		return ACCESN_ERROR;
}

///////////////////////////////////////////////////////////////////////////
// MapPhysicalMemory 
///////////////////////////////////////////////////////////////////////////
/*F*
//  Name:       MapPhysicalMemory 
//  Purpose:    This function maps specified range of physical memory in calling
//              pocesse's address space
//  Context:    Used by Upper level agents (sis modules) to access 
//				system physical memory 
//  Returns:    ACCESN_OK  else error status code
//  Parameters: 
//     
//     startAddress   starting physical address of the  memory to be mapped 
//     addressLength  length of the physical memory to be mapped
//     virtualAddress pointer to the mapped virtual address
//  Notes:      none
*F*/
ACCESN_STATUS
MapPhysicalMemory(int startAddress,int addressLength, int *virtualAddress )
{
	int 				fd,i;
	unsigned int 		length = addressLength;
	off_t 				startpAddress = (off_t)startAddress;
	unsigned int 		diff;
	caddr_t 			startvAddress;

	if ((startAddress == (int) NULL) || (addressLength <= 0))
		return ACCESN_ERROR;


	if ( (fd = open(IMB_DEVICE, O_RDONLY)) < 0) {
		char buf[128];

		sprintf(buf,"%s %s: open(%s) failed\n",
                            __FILE__,__FUNCTION__,IMB_DEVICE);
		perror(buf);
		return ACCESN_ERROR ;
	}

	/* aliging the offset to a page boundary and adjusting the length */
	diff = (int)startpAddress % PAGESIZE;
	startpAddress -= diff;
	length += diff;


	if ( (startvAddress = mmap(	(caddr_t)0, 
								length, 
								PROT_READ, 
								MAP_SHARED, 
								fd, 
								startpAddress
								) ) == (caddr_t)-1)
	{
		char buf[128];

		sprintf(buf,"%s %s: mmap failed\n", __FILE__,__FUNCTION__);
		perror(buf);
		close(fd);
		return ACCESN_ERROR;
	}

	DEBUG("%s: mmap of 0x%x success\n",__FUNCTION__,startpAddress);
#ifdef LINUX_DEBUG_MAX
// don't want this memory dump for normal level of debugging.
// So, I have put it under a stronger debug symbol. mahendra

	for(i=0; i < length; i++)
	{
		printf("0x%x ", (startvAddress[i]));
		if(isascii(startvAddress[i])) {
			printf("%c ", (startvAddress[i]));
		}
	} 
#endif //LINUX_DEBUG_MAX

	*virtualAddress = startvAddress + diff;
	return ACCESN_OK;

}



////////////////////////////////////////////////////////////////////////////
// UnmapPhysicalMemory 
////////////////////////////////////////////////////////////////////////////

/*F*
//  Name:       UnMapPhysicalMemory 
//  Purpose:    This function unmaps the previously mapped physical memory
//  Context:    Used by Upper level agents (sis modules)  
//  Returns:    ACCESN_OK  else error status code
//  Parameters: 
//     
//     addressLength  length of the physical memory to be mapped
//     virtualAddress pointer to the mapped virtual address
//  Notes:      none
*F*/
ACCESN_STATUS
UnmapPhysicalMemory( int virtualAddress, int Length )
{

	unsigned int diff = 0;

	// page align the virtual address and adjust length accordingly 
	
	diff = 	((unsigned int) virtualAddress) % PAGESIZE;
	virtualAddress -= diff;
	Length += diff;

	DEBUG("%s: calling munmap(0x%x,%d)\n",__FUNCTION__,virtualAddress,Length);

	if(munmap((caddr_t)virtualAddress, Length) != 0)
	{
		char buf[128];

		sprintf(buf,"%s %s: munmap failed\n", __FILE__,__FUNCTION__);
		perror(buf);
		return ACCESN_ERROR;

	}

	DEBUG("%s: munmap(0x%x,%d) success\n",__FUNCTION__,virtualAddress,Length);

	return ACCESN_OK;

}

