/*M*
//                    INTEL CORPORATION PROPRIETARY INFORMATION
//      This software is supplied under the terms of a license agreement or
//      nondisclosure agreement with Intel Corporation and may not be copied
//      or disclosed except in accordance with the terms of that agreement.
//            Copyright (c) 2000 Intel Corporation.  All Rights Reserved.
//
//
//  
*M*/


#ifndef	_SMI_H
#define	_SMI_H

#ifndef LINUX

typedef struct {
	int UnixErrNo;
	LONG Win32err;
}ErrorMap;

/*
 * This error map is used for open, close, read, and write
 * system calls. ioctl's pass the NT status block.
 * The driver uses this to map the Win32 API status to a Unix status.
 * A library would used it to map the Unix Status back to Win32 API status.
 */

#define	ERROR_MAP	{ \
	{ ECHILD	, ERROR_GEN_FAILURE		},	\
	{ EBUSY		, ERROR_BUSY			},	\
	{ EBADF		, ERROR_INVALID_HANDLE		},	\
	{ EOVERFLOW	, ERROR_INSUFFICIENT_BUFFER	},	\
	{ EINVAL	, ERROR_INVALID_FUNCTION	},	\
	{ ENOMEM	, ERROR_NOT_ENOUGH_MEMORY	},	\
	{ ENODATA	, ERROR_NO_DATA_DETECTED	},	\
	{ EIO		, ERROR_IO_DEVICE		},	\
	{ ENOENT	, ERROR_FILE_NOT_FOUND		},	\
	{ EINVAL	, ERROR_INVALID_PARAMETER	},	\
	{ EINTR		, ERROR_OPERATION_ABORTED	},	\
	{ EFAULT	, ERROR_READ_FAULT		},	\
	{ EFAULT	, ERROR_WRITE_FAULT		},	\
	{ EACCES	, ERROR_ACCESS_DENIED		},	\
	{ ENOSPC	, ERROR_ALLOTTED_SPACE_EXCEEDED	},	\
	{ ENFILE	, ERROR_TOO_MANY_OPEN_FILES	},	\
	{ EINPROGRESS	, ERROR_IO_PENDING		},	\
	{ E2BIG		, ERROR_MORE_DATA		},	\
	{ 0		, ERROR_SUCCESS			},	\
	} 

#endif /* LINUX */

#define SMI_Version1_00	0x00001000

struct smi {
    DWORD smi_VersionNo;
    DWORD smi_Reserved1;
    DWORD smi_Reserved2;
    LPVOID ntstatus;	/* address of NT status block*/
    LPVOID  lpvInBuffer;        /* address of buffer for input data*/
    DWORD  cbInBuffer;  /* size of input buffer*/
    LPVOID  lpvOutBuffer;       /* address of output buffer*/
    DWORD  cbOutBuffer; /* size of output buffer*/
    LPDWORD  lpcbBytesReturned; /* address of actual bytes of output*/
    LPOVERLAPPED  lpoOverlapped;         /* address of overlapped structure*/
};

#ifdef INKERNEL

#ifndef LINUX

/*
 * This error map is used for open, close, read, and write
 * system calls. ioctl's pass the NT status block.
 * The driver uses this to map the NT status to a Win32 status.
 */

typedef struct {
	LONG Win32err;
	NTSTATUS NTerror;
}NTErrorMap;


#define	NTERROR_MAP	{ \
	{ ERROR_GEN_FAILURE,		STATUS_UNSUCCESSFUL	},	\
	{ ERROR_BUSY,			STATUS_DEVICE_BUSY	},	\
	{ ERROR_INVALID_HANDLE,		STATUS_FILE_CLOSED	},	\
	{ ERROR_INSUFFICIENT_BUFFER,	STATUS_BUFFER_TOO_SMALL	},	\
	{ ERROR_INVALID_FUNCTION,	STATUS_INVALID_DEVICE_REQUEST },\
	{ ERROR_NOT_ENOUGH_MEMORY,	STATUS_INSUFFICIENT_RESOURCES },\
	{ ERROR_NO_DATA_DETECTED,	STATUS_NO_DATA_DETECTED	},	\
	{ ERROR_IO_DEVICE,		STATUS_IO_DEVICE_ERROR	},	\
	{ ERROR_FILE_NOT_FOUND,		STATUS_NO_SUCH_DEVICE	},	\
	{ ERROR_INVALID_PARAMETER,	STATUS_INVALID_PARAMETER},	\
	{ ERROR_ACCESS_DENIED,		STATUS_ACCESS_DENIED	},	\
	{ ERROR_ALLOTTED_SPACE_EXCEEDED,STATUS_ALLOTTED_EXCEEDED},	\
	{ ERROR_TOO_MANY_OPEN_FILES,	STATUS_TOO_MANY_OPEN_FILES},	\
	{ ERROR_OPERATION_ABORTED,	STATUS_CANCELLED	},	\
	{ ERROR_IO_PENDING,		STATUS_PENDING		},	\
	{ ERROR_MORE_DATA,		STATUS_BUFFER_OVERFLOW	},	\
	{ ERROR_SUCCESS,		STATUS_SUCCESS		},	\
	} 


typedef
NTSTATUS
(*PDRIVER_ENTRY)(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	);

typedef
VOID
(*PDRIVER_PFR) (
	IN PDRIVER_OBJECT DriverObject,
	IN int flag
	);

struct NTdrivers {
        NTSTATUS EntryStatus;
        PDRIVER_OBJECT DrvObj;
        PDRIVER_ENTRY DriverEntry;
/* Following line commented because we do not have Power Fail Recovery, corresponding line in Space.c also removed..  3/18/97*/
/*	PDRIVER_PFR DriverPfr;*/
	LONG dependency;	/* index, minor no. of driver dependency */
        PDEVICE_OBJECT DevObj;
	UNICODE_STRING DevStr;
};

#define	NT_EntryReserved	((PDRIVER_OBJECT)-1)
#define	STATUS_NOT_LOADED	((NTSTATUS)-1)
#define	STATUS_LOADING		((NTSTATUS)-2)
#define	SMI_RESERVED_ENTRY	{ STATUS_NOT_LOADED, NT_EntryReserved, \
				NULL, NULL, -1, NULL, {0,0,NULL} }
#define	SMI_AVAILABLE_ENTRY	{ STATUS_NOT_LOADED, NULL, \
				NULL, NULL, -1, NULL, {0,0,NULL} }

/*
 * prototypes in ntddk.c & unixdrv.c 
 */

BOOLEAN
InsertDeviceObject(
	IN PDRIVER_OBJECT DriverObject,
	IN PDEVICE_OBJECT DeviceObject,
	IN PUNICODE_STRING DeviceName OPTIONAL
	);

NTSTATUS 
LookUpDeviceObject(
        IN PUNICODE_STRING  TargetDevice,
        OUT PDEVICE_OBJECT  *AttachedDevice
	);

NTSTATUS
PendingUnixIrp(
    IN PIRP Irp
    );

int
Wstrlen(
	PWSTR s
	);

#endif // LINUX

#ifdef SMITRACE

/*
 * prior_caller(0) returns the calling function.
 * prior_caller(1) returns the callers-caller.
 *      and so on ...
 */
char *
prior_caller(
        int n
	);

struct eventdef {
    int type;
    int numparams;
    char *fmtstr;
} ;

#define	TV_UnixCalls	0x10
#define	TV_IoIrps	0x20
#define	TV_IoCalls	0x30
#define	TV_KeCalls	0x40
#define	TV_KeTimerCalls	0x45
#define	TV_KeLockCalls	0x45
#define	TV_HalCalls	0x50
#define	TV_MmCalls	0x60
#define	TV_RtlCalls	0x70


void SmiTrace (struct eventdef *e, ...);

#endif	/* SMITRACE*/

#else	/* INKERNEL*/

#ifndef STATUS_SUCCESS
typedef struct _IO_STATUS_BLOCK {
    ULONG Status;
    ULONG Information;
} IO_STATUS_BLOCK, *PIO_STATUS_BLOCK;

/*
 * I2C ioctl's return NTStatus codes
 */

#define STATUS_SUCCESS                   (0x00000000U)
#define STATUS_UNSUCCESSFUL              (0xC0000001U)
#define STATUS_DEVICE_BUSY               (0x80000011U)
#define STATUS_PENDING                   (0x00000103U)
#define STATUS_INVALID_PARAMETER         (0xC000000DU)
#define STATUS_INVALID_DEVICE_REQUEST    (0xC0000010U)
#define STATUS_BUFFER_TOO_SMALL          (0xC0000023U)
#define STATUS_FILE_CLOSED               (0xC0000128U)
#define STATUS_INSUFFICIENT_RESOURCES    (0xC000009AU)
#define STATUS_NO_DATA_DETECTED          (0x80000022U)
#define STATUS_NO_SUCH_DEVICE            (0xC000000EU)
#define STATUS_ALLOTTED_EXCEEDED         (0xC000000FU)
#define STATUS_IO_DEVICE_ERROR           (0xC0000185U)
#define STATUS_TOO_MANY_OPEN_FILES       (0xC000011FU)
#define STATUS_ACCESS_DENIED             (0xC0000022U)
#define STATUS_BUFFER_OVERFLOW           (0x80000005U)
#define STATUS_CANCELLED                 (0xC0000120U)

#endif	/* STATUS_SUCCESS*/

#endif	/* INKERNEL*/

#endif	/* _SMI_H*/
